import json

from mcclient.response import SLPResponse, LegacySLPResponse
from mcclient.base_client import BaseClient
from mcclient.encoding.packet import Packet
from mcclient.base_client import DEFAULT_PORT, DEFAULT_TIMEOUT, DEFAULT_PROTO


class SLPClient(BaseClient):
    retries: int

    def __init__(self, host: str, port: int = DEFAULT_PORT, timeout: int = DEFAULT_TIMEOUT, proto: int = DEFAULT_PROTO, srv: bool = True):
        super().__init__(host=host, port=port, timeout=timeout, proto=proto, srv=srv)
        self.retries = 0

    def _status_request(self) -> dict:
        packet = Packet(b"\x00")  # send status request
        self._send(packet)
        res = self._recv()

        if res[0]:
            # if packetloss occured
            if self.retries < 3:
                self.retries += 1
                self._reset()
                return self._status_request()

            else:
                raise Exception("Max retries exceeded.")

        self._close(flush=False)

        res_data = res[2][2:]
        res_str = res_data.decode("utf-8")
        res_dict = json.loads(res_str)
        return res_dict

    def get_status(self) -> SLPResponse:
        self._connect()
        self._handshake()
        res = self._status_request()
        self.retries = 0
        return SLPResponse(self.hostname, self.port, res)


class LegacySLPClient(BaseClient):
    def __init__(self, host: str, port: int = 25565, timeout: int = 5):
        super().__init__(host=host, port=port, timeout=timeout)

    def get_status(self) -> LegacySLPResponse:
        self._connect()
        # legacy status request
        self.sock.send(b"\xFE\x01")
        raw_res = self.sock.recv(1024)
        self._close()

        # remove padding and other headers
        res_bytes = raw_res[3:]
        res_str = res_bytes.decode("UTF-16-be", errors="ignore")
        res_split_str = res_str.split("\x00")
        res = LegacySLPResponse(self.hostname, self.port, res_split_str)
        return res



import socket
import struct

from mcclient.address import Address
from mcclient.encoding.packet import Packet
from mcclient.encoding.varint import VarInt

DEFAULT_HOST = "localhost"
DEFAULT_PORT = 25565
DEFAULT_TIMEOUT = 5
DEFAULT_PROTO = 47


class BaseClient:
    host: str
    hostname: str
    port: int
    sock: socket.socket
    varint: VarInt
    connected: bool | None
    protocol_version: int

    def __init__(self, host: str = DEFAULT_HOST, port: int = DEFAULT_PORT, timeout: int = DEFAULT_TIMEOUT, proto: int = DEFAULT_PROTO, srv: bool = True):
        self.sock = socket.socket(
            socket.AF_INET, socket.SOCK_STREAM)
        self.port = port
        self.varint = VarInt()
        self.connected = False
        self.protocol_version = proto

        self.sock.settimeout(timeout)
        self.get_host(host, port, srv=srv)

    def get_host(self, hostname: str, port: int, srv: bool) -> None:
        addr = Address(hostname)
        self.host, srv_port = addr.get_host(srv)
        self.hostname = hostname
        if srv_port == -1:
            self.port = port

        else:
            self.port = srv_port

    def _connect(self) -> None:
        if self.connected == False:
            self.sock.connect((self.host, self.port))
            self.connected = True

        elif self.connected == None:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.connected = False
            self._connect()

    def _send(self, packet: Packet):
        return self.sock.send(packet.pack())

    def _recv(self) -> tuple[bool, int, bytes]:
        length = self.varint.unpack(self.sock)
        packet_id = self.varint.unpack(self.sock)
        data = self.sock.recv(length)
        if len(data) < length - 4:
            loss = True

        else:
            loss = False
        return loss, packet_id, data

    def _close(self, flush: bool = True):
        if flush:
            self._flush()
        self.sock.close()
        self.connected = None

    def _reset(self) -> None:
        self._close()
        self._connect()
        self._handshake()

    def _flush(self, chunk_size: int = 8192) -> None:
        try:
            self.sock.recv(chunk_size)

        except Exception:
            return

    def implant_socket(self, sock: socket.socket) -> None:
        self.sock = sock
        self.connected = True

    def _handshake(self, next_state: int = 1) -> None:
        packet = Packet(
            b"\x00",  # packet id
            self.varint.pack(self.protocol_version),
            self.hostname,
            struct.pack(">H", self.port),
            self.varint.pack(next_state)  # next state 1 for status request
        )
        self._send(packet)



import struct

from mcclient.encoding.varint import VarInt


class Packet:
    fields: tuple
    varint: VarInt

    def __init__(self, *fields):
        self.fields = fields
        self.varint = VarInt()

    def pack(self) -> bytes:
        packet = b""
        for field in self.fields:
            field = self._encode(field)
            packet += field

        packet = self.varint.pack(len(packet)) + \
            packet  # add the packet length
        return packet

    def _encode(self, data) -> bytes:
        if type(data) == str:
            data = data.encode("utf-8")
            data = self.varint.pack(len(data)) + data

        elif type(data) == bool:
            data = b"\x01" if data else b"\x00"
        return data


class QueryPacket:
    type: int
    session_id: int
    payload: bytes

    def __init__(self, type, session_id, payload):
        self.type = type
        self.session_id = session_id
        self.payload = payload

    def pack(self) -> bytes:
        packet = b"\xFE\xFD"  # query packet padding
        packet += struct.pack("!B", self.type)
        packet += struct.pack('>l', self.session_id)
        packet += self.payload
        return packet



import ipaddress

import dns.resolver


class Address:
    addr: str
    proto: str
    is_ip: bool

    def __init__(self, addr: str, proto: str = "tcp"):
        self.addr = addr
        self.proto = proto
        self.is_ip = self._ip_check(self.addr)

    def get_host(self, srv: bool = True) -> tuple[str, int]:
        if self.is_ip:
            return self.addr, -1

        else:
            return self._resolve_hostname(self.addr, srv)

    def _resolve_hostname(self, hostname: str, srv: bool) -> tuple[str, int]:
        host = self._resolve_a_record(hostname)
        if srv:
            try:
                srv_record = self._mc_srv_lookup(hostname, self.proto)
                return host, srv_record[1]

            except Exception:
                pass

        return host, -1  # -1 = no srv port

    @staticmethod
    def _mc_srv_lookup(hostname: str, proto: str) -> tuple[str, int]:
        srv_prefix = f"_minecraft._{proto}."
        # only use the first srv record returned
        srv_record = dns.resolver.resolve(srv_prefix + hostname, "SRV")[0]
        host = str(srv_record.target).rstrip(".")
        port = int(srv_record.port)
        return host, port

    @staticmethod
    def _resolve_a_record(hostname: str) -> str:
        record = dns.resolver.resolve(hostname, "A")[0]
        return str(record).rstrip(".")

    @staticmethod
    def _ip_check(addr: str) -> bool:
        try:
            ipaddress.ip_address(addr)
            return True
        except ValueError:
            return False
